/*
 *  SPL - The SPL Programming Language
 *  Copyright (C) 2004, 2005  Clifford Wolf <clifford@clifford.at>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  string.c: SPL string implementation
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#include "spl.h"
#include "compat.h"

struct spl_string *spl_string_new(int flags, struct spl_string *left, struct spl_string *right,
		char *text, struct spl_code *code)
{
	struct spl_string *s = calloc(1, sizeof(struct spl_string));

	if (flags & SPL_STRING_NOAUTOGET) {
		s->left = left;
		s->right = right;
	} else {
		s->left = spl_string_get(left);
		s->right = spl_string_get(right);
	}
	s->flags = flags & ~SPL_STRING_NOAUTOGET;
	s->ref_counter = 1;

	s->text = text;
	if (code) {
		s->code = spl_code_get(code);
		s->flags |= SPL_STRING_STATIC;
	}

	return s;
}

struct spl_string *spl_string_printf(int flags, struct spl_string *left, struct spl_string *right,
		const char *fmt, ...)
{
	struct spl_string *s = calloc(1, sizeof(struct spl_string));

	s->left = spl_string_get(left);
	s->right = spl_string_get(right);
	s->flags = flags;
	s->ref_counter = 1;

	if (fmt) {
		va_list ap;
		va_start(ap, fmt);
		my_vasprintf(&s->text, fmt, ap);
		va_end(ap);
	}

	return s;
}

struct spl_string *spl_string_get(struct spl_string *s)
{
	if (s) s->ref_counter++;
	return s;
}

void spl_string_put(struct spl_string *s)
{
	if ( s && --(s->ref_counter) == 0 ) {
		spl_string_put(s->left);
		spl_string_put(s->right);
		spl_code_put(s->code);

		if ( (s->flags & SPL_STRING_STATIC) == 0 && s->text )
			free(s->text);

		free(s);
	}
}

static int spl_string_getlen(struct spl_string *s)
{
	if (!s) return 0;

	int len = 0;
	if (s->left)  len += spl_string_getlen(s->left);
	if (s->text)  len += strlen(s->text);
	if (s->right) len += spl_string_getlen(s->right);
	if ((s->flags & SPL_STRING_DOTCAT) != 0) len++;

	return len;
}

static char *spl_string_write(struct spl_string *s, char *t)
{
	if (!s) return t;

	t = spl_string_write(s->left, t);

	if ((s->flags & SPL_STRING_DOTCAT) != 0) *(t++) = '.';

	if (s->text) {
		int len = strlen(s->text);
		memcpy(t, s->text, len);
		t += len;
	}

	return spl_string_write(s->right, t);
}

char *spl_string(struct spl_string *s)
{
	if (!s) return "";
	if (!s->left && !s->right) return s->text ? s->text : "";

	char *newtext = malloc(spl_string_getlen(s) + 1);
	*(spl_string_write(s, newtext)) = 0;

	spl_string_put(s->left);
	spl_string_put(s->right);
	spl_code_put(s->code);

	s->left = s->right = 0;
	s->code = 0;

	if ( (s->flags & SPL_STRING_STATIC) == 0 && s->text )
		free(s->text);
	s->text = newtext;

	s->flags = s->flags & ~(SPL_STRING_STATIC|SPL_STRING_DOTCAT);

	return newtext;
}

